(require 'cl)
(require 'widget)

(eval-when-compile
  (require 'wid-edit))

(defvar cleite:system:form-values nil)
(defvar cleite:system:second-password nil)


(defun cleite:system:change-password ()
  (interactive)
  "現在接續中のユーザーのパスワードを變更する。
他人のパスワードを變へたい場合は `cleite:system:update-user' を用ゐる事。"
  (block nil
	(let* ((account    (cleite:call 'account:get
									`((user-id . ,cleite:user-id))))
		   (first-one  (read-passwd "Please enter a new password: "))
		   (second-one (read-passwd "Please enter it again: ")))

	  (unless (equal first-one second-one)
		(message "Error: Two passwords you entered are not the same.")
        (return))

      (cleite:call 'account:update
                   `((user-id . ,cleite:user-id)
                     (passwd  . ,first-one)
                     (name    . ,(cdr (assq 'name account)))
                     (is-root . ,(cdr (assq 'is-root account)))))

      (cleite:set-current-password first-one)
      (message "Your password has been properly changed."))))


(defun cleite:system:form-edited (widget key)
  (setq cleite:system:form-values
        (assq-delete-all key cleite:system:form-values))

  (setq cleite:system:form-values
        (cons (cons key (widget-value widget))
              cleite:system:form-values)))


(defun cleite:system:submit-account (mode)
  ;; mode: 'add or 'update
  (block nil
    (let* ((alist   cleite:system:form-values)
           (user-id (cdr (assq 'user-id alist)))
           (passw-1 (cdr (assq 'passwd  alist)))
           (passw-2 cleite:system:second-password)
           (name    (cdr (assq 'name    alist)))
           (is-root (cdr (assq 'is-root alist))))

      (when (= 0 (length user-id))
        (message "Error: User ID must not be empty.")
        (return))

      (unless (equal passw-1 passw-2)
        (message "Error: Two passwords you entered are not the same.")
        (return))

      (when (= 0 (length name))
        (message "Error: User Name must not be empty.")
        (return))

      (when (yes-or-no-p (if (eq 'add mode)
                             "Do you really want to add this account? "
                           "Do you really want to update this account? "))
        (cleite:call (if (eq 'add mode)
                         'account:add
                       'account:update)
                     cleite:system:form-values)
        (kill-buffer (current-buffer))
        (message "Operation succeeded")
        (cleite:feeds)))))


(defun cleite:system:cancel ()
  (when (yes-or-no-p "Do you really want to discard the form? ")
    
    (kill-buffer (current-buffer))
    (cleite:feeds)))


(defun cleite:system:add-account ()
  (interactive)
  (cleite:assert-i-am-root)
  (cleite:system:open-account-form nil))


(defun cleite:system:update-account ()
  (interactive)
  ;; 自分が root でない場合
  ;;   ==> 自分のアカウントしか書換へられないので對象を選ばせる意味は無い。
  ;;
  ;; 自分が root の場合
  ;;   ==> 書換へたいアカウントを一覽から選ばせる。
  ;;
  (labels ((get-account (user-id)
               (cleite:call 'account:get
                            `((user-id . ,user-id)))))

    (if (cleite:root-p)
        (let* ((completion-ignore-case t)
               (user-id (completing-read "User to update: "
                            (mapcar (lambda (account)
                                      (cdr (assq 'user-id account)))
                                    (cleite:call 'account:list))
                            nil  ;; predicate
                            1))) ;; require-match
          (unless (equal "" user-id)
            (cleite:system:open-account-form (get-account user-id))))

      (cleite:system:open-account-form (get-account cleite:user-id)))))


(defun cleite:system:delete-account ()
  (interactive)
  (cleite:assert-i-am-root)
  
  (let* ((completion-ignore-case t)
         (user-id (completing-read "User to delete: "
                      (mapcar (lambda (account)
                                (cdr (assq 'user-id account)))
                              (cleite:call 'account:list))
                      nil  ;; predicate
                      1))) ;; require-match
    (unless (equal "" user-id)
      (cleite:call 'account:delete
                   `((user-id . ,user-id)))
      (message "Operation succeeded"))))


(defun cleite:system:open-account-form (form-values)
  ;; type: 'add or 'update
  (cleite:quit t)
  (switch-to-buffer (if form-values
                        "*Update account*" "*Add new account*"))
  (kill-all-local-variables)
  (make-local-variable 'cleite:system:form-values)
  (make-local-variable 'cleite:system:second-password)
  (use-local-map widget-keymap)
  
  (setq buffer-read-only nil)
  (erase-buffer)

  (let ((all (overlay-lists)))
    ;; 全てのオーバーレイを削除
    (mapcar #'delete-overlay (car all))
    (mapcar #'delete-overlay (cdr all)))
      
  (widget-insert (if form-values
                     "Please fill out this form to update the account.\n"
                   "Please fill out this form to add a new account.\n"))
  (widget-insert "\n")

  (widget-insert "User ID   : ")
  (if form-values
      (widget-insert (cdr (assq 'user-id form-values)))
    
    (widget-create 'editable-field
                   :size   20
                   :notify (lambda (widget &rest ignore)
                             (cleite:system:form-edited widget 'user-id))))
  (widget-insert "\n")

  (widget-insert "User Name : ")
  (widget-create 'editable-field
                 :size   50
                 :value  (if form-values
                             (cdr (assq 'name form-values))
                           "")
                 :notify (lambda (widget &rest ignore)
                           (cleite:system:form-edited widget 'name)))
  (widget-insert "\n")

  (widget-insert "Password  : ")
  (widget-create 'editable-field
                 :size   20
                 :secret ?*
                 :notify (lambda (widget &rest ignore)
                           (cleite:system:form-edited widget 'passwd)))
  (if form-values
      (widget-insert " (Leave this blank if you don't want to change it.)"))
  (widget-insert "\n")
  
  (widget-insert "Password  : ")
  (widget-create 'editable-field
                 :size   20
                 :secret ?*
                 :notify (lambda (widget &rest ignore)
                           (setq cleite:system:second-password
                                 (widget-value widget))))
  (widget-insert " (again)\n")

  (widget-insert "Type      : ")
  (widget-create 'radio-button-choice
                 :indent 12
                 :notify (lambda (widget &rest ignore)
                           (cleite:system:form-edited widget 'is-root))
                 :value  (if form-values
                             (cdr (assq 'is-root form-values))
                           'false)
                 '(item :tag "Normal User" :value false)
                 '(item :tag "Privileged User" :value true))
  (widget-insert "\n")

  (if form-values
      (widget-create 'push-button
                     :notify (lambda (widget &rest ignore)
                               (cleite:system:submit-account 'update))
                     "Update")
    
    (widget-create 'push-button
                   :notify (lambda (widget &rest ignore)
                             (cleite:system:submit-account 'add))
                   "Create"))

  (widget-insert " ")

  (widget-create 'push-button
                 :notify (lambda (widget &rest ignore)
                           (cleite:system:cancel))
                 "Cancel")

  (widget-setup)
  (set-buffer-modified-p nil)
  (goto-char (point-min))
  (widget-forward 1)

  (if form-values
      (setq cleite:system:form-values (cons (cons 'passwd "")
                                            form-values))
    
    (setq cleite:system:form-values
          '((user-id . "")
            (name    . "")
            (passwd  . "")
            (is-root . false))))
  
  (setq cleite:system:second-password ""))


(provide 'cleite-system)